@@ -4,22 +4,108 @@
 
   Author: Pekka Riikonen <priikone@silcnet.org>
 
-  Copyright (C) 1997 - 2001 Pekka Riikonen
+  Copyright (C) 1997 - 2006 Pekka Riikonen
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.
-  
+
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
 */
-/* $Id: silcnet.c,v 1.32.2.2 2005/04/30 15:32:47 priikone Exp $ */
+/* $Id$ */
 
-#include "silcincludes.h"
-#include "silcnet.h"
+#include "silc.h"
+
+/* Returns bound port from listener */
+
+SilcUInt16 *silc_net_listener_get_port(SilcNetListener listener,
+				       SilcUInt32 *port_count)
+{
+  SilcUInt16 *ports;
+  int i;
+
+  ports = silc_calloc(listener->socks_count, sizeof(*ports));
+  if (!ports)
+    return NULL;
+
+  for (i = 0; i < listener->socks_count; i++)
+    ports[i] = silc_net_get_local_port(listener->socks[i]);
+
+  if (port_count)
+    *port_count = listener->socks_count;
+
+  return ports;
+}
+
+/* Return bound IP from listener */
+
+char **silc_net_listener_get_ip(SilcNetListener listener,
+				SilcUInt32 *ip_count)
+{
+  char **ips = NULL, *ip;
+  int i, k;
+
+  ips = silc_calloc(listener->socks_count, sizeof(*ips));
+  if (!ips)
+    return NULL;
+
+  for (i = 0, k = 0; i < listener->socks_count; i++) {
+    if (silc_net_check_local_by_sock(listener->socks[i], NULL, &ip))
+      ips[k++] = ip;
+  }
+
+  if (ip_count)
+    *ip_count = k;
+
+  return ips;
+}
+
+/* Return bound hostname from listener */
+
+char **silc_net_listener_get_hostname(SilcNetListener listener,
+				      SilcUInt32 *hostname_count)
+{
+  char **hs = NULL, *h;
+  int i, k;
+
+  hs = silc_calloc(listener->socks_count, sizeof(*hs));
+  if (!hs)
+    return NULL;
+
+  for (i = 0, k = 0; i < listener->socks_count; i++) {
+    if (silc_net_check_local_by_sock(listener->socks[i], &h, NULL))
+      hs[k++] = h;
+  }
+
+  if (hostname_count)
+    *hostname_count = k;
+
+  return hs;
+}
+
+static const char *silc_net_error[] = {
+  "Ok",
+  "Unknown IP address",
+  "Unknown hostname",
+  "Destination unreachable",
+  "Connection refused",
+  "Connection timeout",
+  "System out of memory",
+  "Unexpected error",
+};
+
+/* Return error as string */
+
+const char *silc_net_get_error_string(SilcNetStatus error)
+{
+  if (error < SILC_NET_OK || error > SILC_NET_ERROR)
+    return "";
+  return silc_net_error[error];
+}
 
 /* Accepts a connection from a particular socket */
 
@@ -37,7 +123,7 @@ int silc_net_set_socket_opt(int sock, int level, int option, int on)
 
 /* Get socket options */
 
-int silc_net_get_socket_opt(int sock, int level, int option, 
+int silc_net_get_socket_opt(int sock, int level, int option,
 			    void *optval, int *opt_len)
 {
   return getsockopt(sock, level, option, optval, opt_len);
@@ -45,12 +131,12 @@ int silc_net_get_socket_opt(int sock, int level, int option,
 
 /* Checks whether IP address sent as argument is valid IPv4 address. */
 
-bool silc_net_is_ip4(const char *addr)
+SilcBool silc_net_is_ip4(const char *addr)
 {
   int count = 0;
 
   while (*addr) {
-    if (*addr != '.' && !isdigit(*addr))
+    if (*addr != '.' && !isdigit((int)*addr))
       return FALSE;
     if (*addr == '.')
       count++;
@@ -59,27 +145,27 @@ bool silc_net_is_ip4(const char *addr)
 
   if (count != 3)
     return FALSE;
-  
+
   return TRUE;
 }
 
 /* Checks whether IP address sent as argument is valid IPv6 address. */
 
-bool silc_net_is_ip6(const char *addr)
+SilcBool silc_net_is_ip6(const char *addr)
 {
   /* XXX does this work with all kinds of IPv6 addresses? */
   while (*addr) {
-    if (*addr != ':' && !isxdigit(*addr))
+    if (*addr != ':' && !isxdigit((int)*addr))
       return FALSE;
     addr++;
   }
-  
+
   return TRUE;
 }
 
 /* Checks whether IP address sent as argument is valid IP address. */
 
-bool silc_net_is_ip(const char *addr)
+SilcBool silc_net_is_ip(const char *addr)
 {
   if (silc_net_is_ip4(addr))
     return TRUE;
@@ -90,7 +176,7 @@ bool silc_net_is_ip(const char *addr)
 typedef struct {
   SilcNetResolveCallback completion;
   void *context;
-  bool prefer_ipv6;
+  SilcBool prefer_ipv6;
   SilcSchedule schedule;
   char *input;
   char *result;
@@ -121,7 +207,7 @@ static void *silc_net_gethostbyname_thread(void *context)
     r->result = strdup(tmp);
 
   silc_schedule_task_add(schedule, 0, silc_net_resolve_completion, r, 0, 1,
-			 SILC_TASK_TIMEOUT, SILC_TASK_PRI_NORMAL);
+			 SILC_TASK_TIMEOUT);
   silc_schedule_wakeup(schedule);
   return NULL;
 }
@@ -138,15 +224,16 @@ static void *silc_net_gethostbyaddr_thread(void *context)
     r->result = strdup(tmp);
 
   silc_schedule_task_add(schedule, 0, silc_net_resolve_completion, r, 0, 1,
-			 SILC_TASK_TIMEOUT, SILC_TASK_PRI_NORMAL);
+			 SILC_TASK_TIMEOUT);
   silc_schedule_wakeup(schedule);
   return NULL;
 }
 
 /* Resolves IP address for hostname. */
 
-bool silc_net_gethostbyname(const char *name, bool prefer_ipv6, char *address, 
-			    SilcUInt32 address_len)
+SilcBool silc_net_gethostbyname(const char *name,
+				SilcBool prefer_ipv6, char *address,
+				SilcUInt32 address_len)
 {
 #ifdef HAVE_IPV6
   struct addrinfo hints, *ai, *tmp, *ip4 = NULL, *ip6 = NULL;
@@ -189,6 +276,14 @@ bool silc_net_gethostbyname(const char *name, bool prefer_ipv6, char *address,
   struct in_addr ip;
   char *tmp;
 
+  if (silc_net_is_ip4(name)) {
+    memset(address, 0, address_len);
+    if (address_len < strlen(name))
+      return FALSE;
+    strncpy(address, name, strlen(name));
+    return TRUE;
+  }
+
   hp = gethostbyname(name);
   if (!hp)
     return FALSE;
@@ -202,14 +297,14 @@ bool silc_net_gethostbyname(const char *name, bool prefer_ipv6, char *address,
   memset(address, 0, address_len);
   strncpy(address, tmp, strlen(tmp));
 #endif
-  
+
   return TRUE;
 }
 
 /* Resolves IP address for hostname async. */
 
-void silc_net_gethostbyname_async(const char *name, 
-				  bool prefer_ipv6,
+void silc_net_gethostbyname_async(const char *name,
+				  SilcBool prefer_ipv6,
 				  SilcSchedule schedule,
 				  SilcNetResolveCallback completion,
 				  void *context)
@@ -227,15 +322,16 @@ void silc_net_gethostbyname_async(const char *name,
 
 /* Resolves hostname by IP address. */
 
-bool silc_net_gethostbyaddr(const char *addr, char *name, SilcUInt32 name_len)
+SilcBool silc_net_gethostbyaddr(const char *addr, char *name,
+				SilcUInt32 name_len)
 {
 #ifdef HAVE_IPV6
   struct addrinfo req, *ai;
-  
+
   memset(&req, 0, sizeof(req));
   req.ai_socktype = SOCK_STREAM;
   req.ai_flags = AI_CANONNAME;
-  
+
   if (getaddrinfo(addr, NULL, &req, &ai))
     return FALSE;
   if (getnameinfo(ai->ai_addr, ai->ai_addrlen, name, name_len, NULL, 0, 0)) {
@@ -245,7 +341,7 @@ bool silc_net_gethostbyaddr(const char *addr, char *name, SilcUInt32 name_len)
   freeaddrinfo(ai);
 #else
   struct hostent *hp;
-  unsigned char a[16];
+  unsigned char a[4];
 
   if (!silc_net_addr2bin(addr, a, sizeof(a)))
     return FALSE;
@@ -258,13 +354,13 @@ bool silc_net_gethostbyaddr(const char *addr, char *name, SilcUInt32 name_len)
   memset(name, 0, name_len);
   strncpy(name, hp->h_name, strlen(hp->h_name));
 #endif
-  
+
   return TRUE;
 }
 
 /* Resolves hostname by IP address async. */
 
-void silc_net_gethostbyaddr_async(const char *addr, 
+void silc_net_gethostbyaddr_async(const char *addr,
 				  SilcSchedule schedule,
 				  SilcNetResolveCallback completion,
 				  void *context)
@@ -279,10 +375,13 @@ void silc_net_gethostbyaddr_async(const char *addr,
   silc_thread_create(silc_net_gethostbyaddr_thread, r, FALSE);
 }
 
+#ifndef SILC_SYMBIAN
+
 /* Performs lookups for remote name and IP address. This peforms reverse
    lookup as well to verify that the IP has FQDN. */
 
-bool silc_net_check_host_by_sock(int sock, char **hostname, char **ip)
+SilcBool silc_net_check_host_by_sock(SilcSocket sock, char **hostname,
+				     char **ip)
 {
   char host[1024];
   int rval, len;
@@ -360,7 +459,8 @@ bool silc_net_check_host_by_sock(int sock, char **hostname, char **ip)
 /* Performs lookups for local name and IP address. This peforms reverse
    lookup as well to verify that the IP has FQDN. */
 
-bool silc_net_check_local_by_sock(int sock, char **hostname, char **ip)
+SilcBool silc_net_check_local_by_sock(SilcSocket sock, char **hostname,
+				      char **ip)
 {
   char host[1024];
   int rval, len;
@@ -437,7 +537,7 @@ bool silc_net_check_local_by_sock(int sock, char **hostname, char **ip)
 
 /* Return remote port by socket. */
 
-SilcUInt16 silc_net_get_remote_port(int sock)
+SilcUInt16 silc_net_get_remote_port(SilcSocket sock)
 {
 #ifdef HAVE_IPV6
   struct sockaddr_storage remote;
@@ -452,7 +552,7 @@ SilcUInt16 silc_net_get_remote_port(int sock)
   if (getnameinfo((struct sockaddr *)&remote, len, NULL, 0, s, sizeof(s),
 		  NI_NUMERICSERV))
     return 0;
-  
+
   return atoi(s);
 #else
   struct sockaddr_in remote;
@@ -469,7 +569,7 @@ SilcUInt16 silc_net_get_remote_port(int sock)
 
 /* Return local port by socket. */
 
-SilcUInt16 silc_net_get_local_port(int sock)
+SilcUInt16 silc_net_get_local_port(SilcSocket sock)
 {
 #ifdef HAVE_IPV6
   struct sockaddr_storage local;
@@ -484,7 +584,7 @@ SilcUInt16 silc_net_get_local_port(int sock)
   if (getnameinfo((struct sockaddr *)&local, len, NULL, 0, s, sizeof(s),
 		  NI_NUMERICSERV))
     return 0;
-  
+
   return atoi(s);
 #else
   struct sockaddr_in local;
@@ -498,6 +598,7 @@ SilcUInt16 silc_net_get_local_port(int sock)
   return ntohs(local.sin_port);
 #endif
 }
+#endif /* !SILC_SYMBIAN */
 
 /* Return name of localhost. */
 
